
#include <nokix/all.h>
#include <stdarg.h>

/*
** timers.c (9.07.2005)
**
** timers subsystem for own patches and applications
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2.1 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

//#define DEBUG

struct node
{
 struct node *next;		/* pointer to next timer in our timers list */
 void *todo;			/* pointer to data used to generate action after
						   timer reaches 0, also a unique value describing
						   a timer */
 UWORD type;			/* type of data pointed by 'todo' */
 UWORD base;			/* hardware timer state during start */
 UWORD time;			/* number of ticks */
};

#define CUT_WORD(x) ((((ULONG)(x))<<16)>>16)

/********************************************************/

ULONG get_timer(void);

/********************************************************/

extern rexx_uint timer_id;	/* ID of system timer, our timers use it */

struct node *list;			/* dynamic list of timers */

/********************************************************/

#ifdef DEBUG
void dump_list(void)
{
 struct node *n=list;
 int i=1;
 LONG t;

 while(n)
 {
  t=n->time-CUT_WORD(get_timer()-n->base);
  if(t<0) t=0;
  printf("Timers list: %d. todo=0x%lX, type=%ld, time=%ld, left=%ld {%08lX %08lX}\n",i,n->todo,n->type,n->time,t,((ULONG *)n->todo)[0],((ULONG *)n->todo)[1]);
  n=n->next;
  i++;
 }
}
#endif

/********************************************************/

void own_timer_start(const void *todo, unsigned type, unsigned time)
{
 struct node *n, *p=NULL;
 UWORD base, t;

// disable_irq();

 if(time==0) time=1;				/* min */
 else if(time>0xffff) time=0xffff;	/* max (~8:44) */

 own_timer_abort(todo); /* remove if it's already there */

#ifdef DEBUG
 printf("Start timer: todo=0x%lX, type=%ld, time=%ld {%08lX %08lX}\n",todo,type,time,((ULONG *)todo)[0],((ULONG *)todo)[1]);
#endif

 base=get_timer();
 n=list;
 while(n) /* find suitable place for our timer in list */
 {
  t=(n->base>base) ? 0x10000 : 0;
  if(t+base+time<n->base+n->time) break;
  p=n; n=n->next;
 }
 
 n=malloc(sizeof(struct node));
 n->todo=(void *)todo;
 n->type=type;
 n->base=base;
 n->time=time;

 if(p) /* somewhere in the middle */
 {
  n->next=p->next;
  p->next=n;
 }
 else /* first */
 {
  n->next=list;
  list=n;
  start_timer(timer_id,time);
 }

#ifdef DEBUG
 dump_list();
#endif

// enable_irq();
}

/********************************************************/

int own_timer_query(const void *todo)
{
 LONG t=TE_TIMER_NOT_FOUND;
 struct node *n=list;

// disable_irq();

 if(list)
 {
  while(n) /* find our timer */
  {
   if(n->todo==todo)
   {
	t=n->time-CUT_WORD(get_timer()-n->base);
	if(t<0) t=0;
	break;
   }
   n=n->next;
  }
 }

// enable_irq();
#ifdef DEBUG
 printf("Query timer: todo=0x%lX [time=%ld]\n",todo,t);
#endif
#ifdef DEBUG
 dump_list();
#endif

 return t;
}

/********************************************************/

int own_timer_abort(const void *todo)
{
 LONG t=TE_TIMER_NOT_FOUND;
 struct node *n=list, *p=NULL;

// disable_irq();
#ifdef DEBUG
 dump_list();
#endif

 if(list)
 {
  while(n) /* find our timer */
  {
   if(n->todo==todo) break;
   p=n; n=n->next;
  }
  if(n) /* if found */
  {
   if(p) p->next=n->next; /* somewhere in the middle */
   else /* first */
   {
	if((list=n->next)) /* if we sill have timers */
	{
     t=list->time-CUT_WORD(get_timer()-list->base);
	 if(t<0) t=0;
     start_timer(timer_id,t);
    }
	else abort_timer(timer_id); /* this was last */
   }
   t=n->time-CUT_WORD(get_timer()-n->base);
   if(t<0) t=0;
   free(n);
   free((void *)todo);
  }
 }

// enable_irq();
#ifdef DEBUG
 printf("Abort timer: todo=0x%lX [time=%ld]\n",todo,t);
#endif

return t;
}

/********************************************************/
/* called when UI task timer message arrives (called in asm.s file) */

void handle_timer(struct node *tim, void *todo)
{
#ifdef DEBUG
 printf("Handle timer: 0x%lX; todo=0x%lX\n",tim,todo);
#endif
#ifdef DEBUG
 dump_list();
#endif

 /* we can't handle timer if it was removed in the meantime */
 if((tim!=list)||(todo!=list->todo)) return;

 /* remove timer */
 list=tim->next;

 /* init next timer */
 if(list)
 {
  LONG t=list->time-CUT_WORD(get_timer()-list->base);
  if(t<=0) handle_timer(list,list->todo); /* timer expired! handle it immediately! */
  else start_timer(timer_id,t);
 }

 { /* handle the job */
  ULONG *t=(ULONG *)tim->todo;
  
#ifdef DEBUG
 printf("Handle timer: todo=0x%lX, type=%ld {%08lX %08lX}\n",t,tim->type,t[0],t[1]);
#endif

  switch(tim->type)
  {
   case TT_MESSAGE:
   {
    send_message(t[0],t[1],t[2],t[3]);
    break;
   }
   case TT_FUNCTION:
   {
    void (*func)(void)=(void (*)(void))t;
    func();
    break;
   }
   case TT_MESSAGES:
   {
    while(t[0]!=TMSG_END)
    {
     send_message(t[0],t[1],t[2],t[3]);
     t+=((t[0]>>14)&3)+1; /* t[0]>>14 = number of args */
    }
    break;
   }
   case TT_TASK_MESSAGE:
   {
    send_task_message(t[0],(void *)t[1]);
    //break;
   }
  }
 }

 /* free the timer struct */
 free(tim->todo);
 free(tim);
}

/********************************************************/

void *own_timer_create(unsigned time, unsigned type, ...)
{
 va_list vl;
 void *todo=NULL;
 
#ifdef DEBUG
 printf("Create timer: time=%ld, type=%ld\n",time,type);
#endif
#ifdef DEBUG
 dump_list();
#endif

 va_start(vl,type);
 switch(type)
 {
  case TT_MESSAGE:
  {
   ULONG msg=va_arg(vl,ULONG);
   UBYTE argc=(msg>>14)&3;
   ULONG *t;

   t=malloc(4+(4*argc));
   t[0]=msg;
   if(argc>=1) t[1]=va_arg(vl,ULONG);
   if(argc>=2) t[2]=va_arg(vl,ULONG);
   if(argc==3) t[3]=va_arg(vl,ULONG);

   todo=t;
   break;
  }
  case TT_FUNCTION:
  {
   todo=malloc(4);
   *((ULONG *)todo)=va_arg(vl,ULONG);
   break;
  }
  case TT_MESSAGES:
  {
   ULONG msg,i=0;
   UBYTE argc;
   ULONG *t=NULL;

   do
   {
    msg=va_arg(vl,ULONG);
    argc=(msg>>14)&3;
    t=realloc(t,(i+1+argc)*sizeof(ULONG));
    t[i]=msg;
    if(argc>=1) t[i+1]=va_arg(vl,ULONG);
    if(argc>=2) t[i+2]=va_arg(vl,ULONG);
    if(argc==3) t[i+3]=va_arg(vl,ULONG);
    i+=1+argc;
   }
   while(msg!=TMSG_END);

   todo=t;
   break;
  }
  case TT_TASK_MESSAGE:
  {
   todo=malloc(8);
   ((ULONG *)todo)[0]=va_arg(vl,ULONG);
   ((ULONG *)todo)[1]=va_arg(vl,ULONG);
   break;
  }
  case TT_NOP:
  {
   todo=malloc(1); // dummy
   //break;
  }
 }
 va_end(vl);

 if(todo) own_timer_start(todo,type,time);
 return todo;
}
